/* 
 * verse_gimp.c is part of the Verse Gimp Plugin.
 * Copyright (C) 2003 Brecht Van Lommel <blendix@pandora.be>
 *
 * Verse Gimp Plugin is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * Verse Gimp Plugin is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Verse Gimp Plugin; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#define GTK_ENABLE_BROKEN /* for the textview */

#include <verse.h>
#undef DBL_MAX
#undef TRUE
#undef FALSE
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <gtk/gtk.h>
#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>

#include "verse_gimp.h"
#include "verse_connection.h"

static gboolean create_gimp_image_from_verse(i_bitmap_t *bitmap);
static void cb_gui_connect(GtkWidget *widget, gpointer data);
static void cb_gui_disconnect(GtkWidget *widget, gpointer data);
static void cb_gui_upload(GtkWidget *widget, gpointer data);
static void cb_gui_download(GtkWidget *widget, gpointer data);
static void cb_gui_unlink(GtkWidget *widget, gpointer data);
static void cb_gui_mode(GtkWidget *widget, gpointer data);
static void gui_refresh_gimp_list();
static void nice_quit();
static gboolean gui_quit(gpointer user);
static void gui_destroy();
static int gimp_images_update();

typedef struct
{
	/* Widgets */
	GtkWidget *dialog;
	GtkWidget *notebook;

	GtkWidget *state_label;
	GtkWidget *conn;
	GtkWidget *disconn;

	GtkWidget *table_server;
	GtkWidget *address;
	GtkObject *speedadj;
	GtkWidget *uname;
	GtkWidget *pass;

	GtkListStore *gmplist;
	GtkWidget *gmptree;

	GtkListStore *vrslist;
	GtkWidget *vrstree;
	GList *bitmaps;
	GArray *upload_queue;

	/* Program State */
	gboolean connected;
	gboolean connecting;
	gboolean disconnecting;

	/* Some Data to Remember */
	guint verse_timeout;
	guint gimp_timeout;
	VSession *session;
	guint32 avatar;
	uint tiles_sent;
} verse_gimp_t;

static verse_gimp_t *vg;

/* ### Pixel Operations ### */

void put_layer_data(i_layer_t *layer, uint16 tile_x, uint16 tile_y, VNBTile *data, VNBLayerType type)
{
	uint i, j, gtile = (tile_y / (NTILE_SIZE / VN_B_TILE_SIZE)) * NTILES(layer->parent->width) + (tile_x / (NTILE_SIZE / VN_B_TILE_SIZE));
	i_tile_t *tile = &(layer->data[gtile]);
	/*layer->parent->redraw_pending = TRUE;*/
	uint8 *dest = &(tile->data[((tile_y % (NTILE_SIZE / VN_B_TILE_SIZE)) * (NTILE_SIZE / VN_B_TILE_SIZE) + (tile_x % (NTILE_SIZE / VN_B_TILE_SIZE))) * ONE_TILE]);
	
	if (type == VN_B_LAYER_UINT1) {
		for (i = 0; i < VN_B_TILE_SIZE; i++)
			for (j = 0; j < VN_B_TILE_SIZE; j++)
				dest[i + j*VN_B_TILE_SIZE] = (data->vuint1[j] & (1 << i))*255;
	} else if (type == VN_B_LAYER_UINT8) {
		for (i = 0; i < ONE_TILE; i++)
			dest[i] = data->vuint8[i];
	} else if (type == VN_B_LAYER_UINT16) {
		for (i = 0; i < ONE_TILE; i++)
			dest[i] = data->vuint16[i] / 255;
	} else if (type == VN_B_LAYER_REAL32) {
		for (i = 0; i < ONE_TILE; i++)
			dest[i] = data->vreal32[i] * 255;
	} else if (type == VN_B_LAYER_REAL64) {
		for (i = 0; i < ONE_TILE; i++)
			dest[i] = data->vreal64[i] * 255;
	}
	gtile++; /* in a g_array 0 means we are at the end of the array, so, to
	            be able to have a 0 tile we do everything +1 */
	for (i = 0; (j = g_array_index(layer->parent->redraw_queue, guint, i)) != 0; i++) {
		if (j == gtile)
			return;
	}
	g_array_append_val(layer->parent->redraw_queue, gtile);

}

void redraw_tile_data(i_bitmap_t *bitmap, uint tile_x, uint tile_y)
{
	GimpDrawable *drawable;
	GimpTile *tile;
	uint8 *new_data;
	guint i, offset;
	guchar tx, ty, x, y, tile_w, tile_h, j;
	i_layer_t *verse_layers[4];
	GList *iter;
	i_layer_t *layer;
	i_tile_t *ntile;

	drawable = gimp_drawable_get(bitmap->gimp_layer_id);
	tile = gimp_drawable_get_tile(drawable, TRUE, tile_y, tile_x);
	gimp_tile_ref(tile);

	for (i = 0; i < 4; i++)
		verse_layers[i] = NULL;
	for (iter = bitmap->layers; iter != NULL; iter = g_list_next(iter)) {
		layer = (i_layer_t*)iter->data;
		/* XXX layer->data != NULL is temp fix, tot resizen mogelijk is */
		if (layer->gimp_channel >= 0 && layer->data != NULL)
			verse_layers[layer->gimp_channel] = layer;
	}

	for (tx = 0; tx < TILES(tile->ewidth); tx++) {
		for (ty = 0; ty < TILES(tile->eheight); ty++) {
			tile_w = ((tx + 1) * VN_B_TILE_SIZE > tile->ewidth) ? tile->ewidth % VN_B_TILE_SIZE : VN_B_TILE_SIZE;
			tile_h = ((ty + 1) * VN_B_TILE_SIZE > tile->eheight) ? tile->eheight % VN_B_TILE_SIZE : VN_B_TILE_SIZE;
			for (j = 0; j < 4; j++) {
				if (verse_layers[j] != NULL) {
					ntile = &(verse_layers[j]->data[tile_y * NTILES(bitmap->width) + tile_x]);
					new_data = &(ntile->data[(ty * TILES(ntile->w) + tx) * ONE_TILE]);
					for (y = 0, i = 0; y < tile_h; y++) {
						offset = ((y + ty * VN_B_TILE_SIZE) * tile->ewidth + tx * VN_B_TILE_SIZE) * tile->bpp + j;
						i = y * VN_B_TILE_SIZE;
						for (x = 0; x < tile_w; x++, offset += tile->bpp, i++)
							tile->data[offset] = new_data[i];
					}
				} else {
					/* make sure the image is always visible, even without an
					   alpha layer */
					if (j == 3) {
						for (y = 0, i = 0; y < tile_h; y++) {
							offset = ((y + ty * VN_B_TILE_SIZE) * tile->ewidth + tx * VN_B_TILE_SIZE) * tile->bpp + j;
							for (x = 0; x < tile_w; x++, offset += tile->bpp, i++)
								tile->data[offset] = 255;
						}
					} else {
						for (y = 0, i = 0; y < tile_h; y++) {
							offset = ((y + ty * VN_B_TILE_SIZE) * tile->ewidth + tx * VN_B_TILE_SIZE) * tile->bpp + j;
							for (x = 0; x < tile_w; x++, offset += tile->bpp, i++)
								tile->data[offset] = 0;
						}
					}
				}
			}
		}
	}
	tile->dirty = TRUE;
	gimp_tile_flush(tile);
	gimp_drawable_merge_shadow(drawable->drawable_id, FALSE);
	gimp_drawable_update(bitmap->gimp_layer_id, NTILE_SIZE * tile_x, NTILE_SIZE * tile_y, tile->ewidth, tile->eheight);
	/* FALSE means we don't want to add this to the undo stack */
	gimp_tile_unref(tile, FALSE);
	gimp_drawable_detach(drawable);
	gimp_displays_flush();
}

void update_tile_data(i_bitmap_t *bitmap, uint tile_x, uint tile_y)
{
	GimpDrawable *drawable;
	GimpTile *tile;
	uint8 *new_data, *old_data;
	gboolean data_is_equal;
	guint i, offset;
	guchar tx, ty, x, y, tile_w, tile_h, j;
	i_layer_t *verse_layers[4];
	GList *iter;
	i_layer_t *layer;
	i_tile_t *ntile;
	uint sent = 0;
	VNBTile vtile;

	drawable = gimp_drawable_get(bitmap->gimp_layer_id);
	tile = gimp_drawable_get_tile(drawable, FALSE, tile_y, tile_x);
	gimp_tile_ref(tile);

	for (i = 0; i < 4; i++)
		verse_layers[i] = NULL;
	for (iter = bitmap->layers; iter != NULL; iter = g_list_next(iter)) {
		layer = (i_layer_t*)iter->data;
		/* XXX layer->data != NULL is temp fix, tot resizen mogelijk is */
		if (layer->gimp_channel >= 0/* && layer->data != NULL && layer->data != 0*/)
			verse_layers[layer->gimp_channel] = layer;
	}

	new_data = g_malloc(ONE_TILE * sizeof(uint8));
	for (tx = 0; tx < TILES(tile->ewidth); tx++) {
		for (ty = 0; ty < TILES(tile->eheight); ty++) {
			tile_w = ((tx + 1) * VN_B_TILE_SIZE > tile->ewidth) ? tile->ewidth % VN_B_TILE_SIZE : VN_B_TILE_SIZE;
			tile_h = ((ty + 1) * VN_B_TILE_SIZE > tile->eheight) ? tile->eheight % VN_B_TILE_SIZE : VN_B_TILE_SIZE;
			for (j = 0; j < 4; j++) {
				if (verse_layers[j] != NULL) {
					ntile = &(verse_layers[j]->data[tile_y * NTILES(bitmap->width) + tile_x]);
					old_data = &(ntile->data[(ty * TILES(ntile->w) + tx) * ONE_TILE]);
					data_is_equal = TRUE;
					for (y = 0, i = 0; y < tile_h; y++) {
						offset = ((y + ty * VN_B_TILE_SIZE) * tile->ewidth + tx * VN_B_TILE_SIZE) * tile->bpp + j;
						for (x = 0; x < tile_w; x++, offset += tile->bpp, i++) {
							new_data[i] = tile->data[offset];
							if (data_is_equal)
								data_is_equal = new_data[i] == old_data[i];
						}
					}
					if(!data_is_equal) {
						memcpy(old_data, new_data, sizeof(uint8) * tile_w * tile_h);
						memcpy(vtile.vuint8, new_data, sizeof(uint8) * tile_w * tile_h);
						verse_send_b_tile_set(bitmap->verse_node_id, verse_layers[j]->verse_layer_id, tile_x * NTILE_SIZE / VN_B_TILE_SIZE + tx, tile_y * NTILE_SIZE / VN_B_TILE_SIZE + ty, 0, NATIVE_LAYER_TYPE, &vtile);
						sent++;
					}
				}
			}
		}
	}
	g_free(new_data);
	gimp_tile_unref(tile, FALSE);
	gimp_drawable_detach(drawable);
}

static gboolean check_gimp_changes(i_bitmap_t *bitmap)
{
	gint32 *images, *layers;
	gint num_images, num_layers;
	gboolean image_found, layer_found;
	gint i;

	/* Check if image still exists */
	images = gimp_image_list(&num_images);
	image_found = FALSE;
	for (i = 0; i < num_images && !image_found; i++)
		if (bitmap->gimp_image_id == images[i])
			image_found = TRUE;
	g_free(images);
	if (!image_found)
		return FALSE;

	/* Check if layer still exists */
	layers = gimp_image_get_layers(bitmap->gimp_image_id, &num_layers);
	layer_found = FALSE;
	for (i = 0; i < num_layers && !layer_found; i++)
		if (bitmap->gimp_layer_id == layers[i])
				layer_found = TRUE;
	if (!layer_found)
		return FALSE;

	/* Check if image is still rgb and alpha */
	if (!gimp_drawable_has_alpha(bitmap->gimp_layer_id))
		return FALSE;

	if (!gimp_drawable_is_rgb(bitmap->gimp_layer_id))
		return FALSE;
		
	/* Check if width and height haven't changed */
	if (gimp_drawable_width(bitmap->gimp_layer_id) != bitmap->width ||
	    gimp_drawable_height(bitmap->gimp_layer_id) != bitmap->height)
		return FALSE;

	return TRUE;
}

static gboolean check_and_fix_gimp_changes(i_bitmap_t *bitmap)
{
	gint32 *images, *layers;
	gint num_images, num_layers;
	gboolean image_found, layer_found;
	gint i;

	/* Check if image still exists */
	images = gimp_image_list(&num_images);
	image_found = FALSE;
	for (i = 0; i < num_images && !image_found; i++)
		if (bitmap->gimp_image_id == images[i])
			image_found = TRUE;
	g_free(images);
	if (!image_found)
		return FALSE;

	/* Check if layer still exists, if not, use next layer */
	layers = gimp_image_get_layers(bitmap->gimp_image_id, &num_layers);
	layer_found = FALSE;
	for (i = 0; i < num_layers && !layer_found; i++)
		if (bitmap->gimp_layer_id == layers[i])
				layer_found = TRUE;
	if (!layer_found) {
		if (num_layers > 0) {
			bitmap->gimp_layer_id = layers[0];
			g_free(layers);
		} else {
			g_free(layers);
			return FALSE;
		}
	}

	/* Check if image is still rgb and alpha */
	if (!gimp_drawable_has_alpha(bitmap->gimp_layer_id))
		gimp_layer_add_alpha(bitmap->gimp_layer_id);

	if (!gimp_drawable_is_rgb(bitmap->gimp_layer_id))
		gimp_image_convert_rgb(bitmap->gimp_image_id);
		
	/* Check if width and height haven't changed. FIXME resize??? */
	if (gimp_drawable_width(bitmap->gimp_layer_id) != bitmap->width ||
	    gimp_drawable_height(bitmap->gimp_layer_id) != bitmap->height)
		return FALSE;
	return TRUE;
}

static int gimp_images_update()
{
	GList *i;
	i_bitmap_t *bitmap;
	uint redraw_tile;
	/*clock_t starttime;
	starttime = clock();*/

	vg->tiles_sent = 0;
	for (i = vg->bitmaps; i != NULL; i = g_list_next(i)) {
		bitmap = (i_bitmap_t*)i->data;
		if (bitmap != NULL && bitmap->width > 0 && bitmap->height > 0 &&
		    bitmap->layers != NULL) {
			if (bitmap->upload) {
				if (!check_and_fix_gimp_changes(bitmap))
					unlink_bitmap(bitmap);
				else {
					update_tile_data(bitmap, bitmap->current_tile % NTILES(bitmap->width), bitmap->current_tile / (NTILES(bitmap->width)));
					bitmap->current_tile++;
					if (bitmap->current_tile >= NTILES(bitmap->width) * NTILES(bitmap->height))
						bitmap->current_tile = 0;
					vrs_callback_update();
				}
			} else if (bitmap->download) {
				if (!check_gimp_changes(bitmap))
					unlink_bitmap(bitmap);
				else if (bitmap->layers != NULL/* && bitmap->redraw_pending*/) {
					if ((redraw_tile = pop_redraw_queue(bitmap)) != 0) {
						/* XXX we hope redraw_tile still exists */
						redraw_tile--;
						gimp_image_undo_freeze(bitmap->gimp_image_id);
						redraw_tile_data(bitmap, redraw_tile % NTILES(bitmap->width), redraw_tile / NTILES(bitmap->width));
						/*bitmap->redraw_pending = FALSE;*/
						gimp_image_undo_thaw(bitmap->gimp_image_id);
					}
				}
			}
		}
	}
	/* FIXME */
	/*vrs_callback_update();
	starttime = clock() - starttime;
	printf("Elapsed time = %d.%02d seconds\n",(int) (starttime/CLK_TCK), (int) (starttime % CLK_TCK));*/
	return TRUE;
}

/* ### Variable Acces ### */

void gui_set_avatar(guint32 avatar) { vg->avatar = avatar; }
guint32 gui_get_avatar() { return vg->avatar; }
guint32 gui_get_verse_timeout() { return vg->verse_timeout; }
guint gui_get_verse_update_interval() { return 100; }

gint32 pop_id_upload_queue()
{
	/* since this is a 0-terminated array, it will return 0 in case there
	   are no elements. this is ok, since we know there are no gimp images
	   with id=0. */
	gint32 i;
	if ((i = g_array_index(vg->upload_queue, gint32, 0)) != 0) {
		g_array_remove_index(vg->upload_queue, 0);
		return i;
	} else
		return 0;
}

guint pop_redraw_queue(i_bitmap_t *bitmap)
{
	/* since this is a 0-terminated array, it will return 0 in case there
	   are no elements. this is ok, since we know there are no gimp images
	   with id=0. */
	guint i;
	if ((i = g_array_index(bitmap->redraw_queue, guint, 0)) != 0) {
		g_array_remove_index(bitmap->redraw_queue, 0);
		return i;
	} else
		return 0;
}

i_bitmap_t *get_bitmap(VNodeID id)
{
	GList *i;
	i_bitmap_t *bitmap;
	for (i = vg->bitmaps; i != NULL; i = g_list_next(i))
		if ((bitmap = (i_bitmap_t*)i->data) != NULL)
			if (bitmap->verse_node_id == id)
				return bitmap;
	return NULL;
}

i_bitmap_t *get_bitmap_verse_id(gint32 id)
{
	GList *i;
	i_bitmap_t *bitmap;
	for (i = vg->bitmaps; i != NULL; i = g_list_next(i))
		if ((bitmap = (i_bitmap_t*)i->data) != NULL)
			if (bitmap->gimp_image_id == id)
				return bitmap;
	return NULL;
}


i_layer_t *get_layer(i_bitmap_t *bitmap, VLayerID layer_id)
{
	GList *i;
	i_layer_t *layer;
	for (i = bitmap->layers; i != NULL; i = g_list_next(i))
		if ((layer = (i_layer_t*)i->data) != NULL)
			if (layer->verse_layer_id == layer_id)
				return layer;
	return NULL;
}

static gboolean create_gimp_image_from_verse(i_bitmap_t *bitmap)
{
	if (bitmap->gimp_image_id > 0)
		return FALSE;
	bitmap->gimp_image_id = gimp_image_new(bitmap->width, bitmap->height, GIMP_RGB);
	bitmap->gimp_layer_id = gimp_layer_new (bitmap->gimp_image_id,
	                           "Background",
	                           bitmap->width,
	                           bitmap->height,
	                           GIMP_RGBA_IMAGE, 100, GIMP_NORMAL_MODE);
	gimp_image_add_layer (bitmap->gimp_image_id, bitmap->gimp_layer_id, 0);
	gimp_drawable_fill(bitmap->gimp_layer_id, GIMP_WHITE_FILL);
	gimp_display_new(bitmap->gimp_image_id);
	return TRUE;
}

i_bitmap_t *add_bitmap_from_verse(VNodeID node_id)
{
	i_bitmap_t *bitmap;
	if (get_bitmap(node_id) != NULL) 
		return NULL;
	bitmap = g_malloc(sizeof *bitmap);
	bitmap->gimp_image_id = 0;
	bitmap->gimp_layer_id = 0;
	bitmap->verse_node_id = node_id;
	bitmap->upload = FALSE;
	bitmap->creation_pending = FALSE;
	/*bitmap->redraw_pending = FALSE;*/
	bitmap->download = FALSE;
	bitmap->current_tile = 0;
	bitmap->width = 0;
	bitmap->height = 0;
	bitmap->layers = NULL;
	bitmap->name = NULL;
	bitmap->redraw_queue = g_array_new(TRUE, FALSE, sizeof(guint));
	vg->bitmaps = g_list_append(vg->bitmaps, bitmap);

	return bitmap;
}

gboolean set_bitmap_from_gimp(i_bitmap_t *bitmap, gint32 gimp_image_id)
{
	gchar *name;
	gint32 *layers;
	gint num_layers;

	bitmap->gimp_image_id = gimp_image_id;
	bitmap->upload = TRUE;
	bitmap->download = FALSE;
	bitmap->creation_pending = TRUE;

	g_free(bitmap->name);
	bitmap->name = NULL;
	name = gimp_image_get_name(gimp_image_id);
	bitmap->name = g_strdup(name);
	g_free(name);

	layers = gimp_image_get_layers(bitmap->gimp_image_id, &num_layers);
	if (num_layers < 1) {
		gimp_message("The image must at least have one layer to upload.");
		return FALSE;
	} else if (num_layers > 1)
		gimp_message("The image has multiple layers. Only the topmost layer will be used.");
	bitmap->gimp_layer_id = layers[0];
	g_free(layers);

	bitmap->width = gimp_drawable_width(bitmap->gimp_layer_id);
	bitmap->height = gimp_drawable_height(bitmap->gimp_layer_id);

	if (!gimp_drawable_is_rgb(bitmap->gimp_layer_id)) {
		GtkWidget *dialog;
		gint result;
		dialog = gtk_message_dialog_new (GTK_WINDOW(vg->dialog),
		                                 GTK_DIALOG_DESTROY_WITH_PARENT,
		                                 GTK_MESSAGE_QUESTION,
		                                 GTK_BUTTONS_OK_CANCEL,
		"The image you are about to upload is an Indexed or Gray image. The image will be converted to RGB before uploading. Proceed?");
		result = gtk_dialog_run(GTK_DIALOG(dialog));
		gtk_widget_destroy(dialog);
		if (result != GTK_RESPONSE_OK)
			return FALSE;
		gimp_image_convert_rgb(bitmap->gimp_image_id);
	}

	if (!gimp_drawable_has_alpha(bitmap->gimp_layer_id))
		gimp_layer_add_alpha(bitmap->gimp_layer_id);

	return TRUE;
}

void allocate_layer_data(i_bitmap_t *bitmap, i_layer_t *layer) {
	uint x, y;
	i_tile_t *tile;

	layer->data = NULL;
	if (bitmap->width * bitmap->height > 0) {
		layer->data = g_malloc(sizeof(i_tile_t) * NTILES(bitmap->width) * NTILES(bitmap->height));
		for (x = 0; x < NTILES(bitmap->width); x++) {
			for (y = 0; y < NTILES(bitmap->height); y++) {
				tile = &(layer->data[y * NTILES(bitmap->width) + x]);
				tile->w = /*(x + 1) * NTILE_SIZE > bitmap->width ? bitmap->width % NTILE_SIZE :*/ NTILE_SIZE;
				tile->h = /*(y + 1) * NTILE_SIZE > bitmap->height ? bitmap->height % NTILE_SIZE :*/ NTILE_SIZE;
				tile->dirty = FALSE;
				tile->data = g_malloc0(sizeof(uint8) * TILES(tile->w) * TILES(tile->h) * ONE_TILE);
			}
		}
	}
}

void free_layer_data(i_bitmap_t *bitmap, i_layer_t *layer) {
	uint x, y;
	i_tile_t *tile;

	if (layer->data != NULL && bitmap->width * bitmap->height > 0) {
		for (x = 0; x < NTILES(bitmap->width); x++) {
			for (y = 0; y < NTILES(bitmap->height); y++) {
				tile = &(layer->data[y * NTILES(bitmap->width) + x]);
				g_free(tile->data);
			}
		}
		g_free(layer->data);
	}
}

i_layer_t *add_layer_from_verse(i_bitmap_t *bitmap, VLayerID layer_id, char *name, VNBLayerType type)
{
	i_layer_t *layer;
	gint channel;
	GList *iter;

	/* determine the layer's channel from the name */
	if (strcmp(name, "col_r") == 0)
		channel = 0;
	else if (strcmp(name, "col_g") == 0)
		channel = 1;
	else if (strcmp(name, "col_b") == 0)
		channel = 2;
	else if (strcmp(name, "alpha") == 0)
		channel = 3;
	else
		return NULL;

	if (get_layer(bitmap, layer_id) != NULL)
		return NULL;
	layer = g_malloc(sizeof *layer);
	layer->parent = bitmap;
	layer->name = g_strdup(name);
	layer->verse_layer_id = layer_id;
	layer->verse_layer_type = type;
	layer->gimp_channel = 0;

	allocate_layer_data(bitmap, layer);


	layer->gimp_channel = channel;

	/* make sure there is only one layer per channel */
	for (iter = bitmap->layers; iter != NULL; iter = g_list_next(iter))
		if (((i_layer_t*)iter->data)->gimp_channel == channel)
			layer->gimp_channel = -1;

	bitmap->layers = g_list_append(bitmap->layers, layer);
	return layer;
}

void resize_bitmap(i_bitmap_t *bitmap, uint new_width, uint new_height)
{
	GList *i;
	i_layer_t *layer;
	uint old_width = bitmap->width, old_height = bitmap->height;
	if (bitmap->layers != NULL) {
		for (i = bitmap->layers; i != NULL; i = g_list_next(i)) {
			layer = (i_layer_t*)i->data;
			if (layer->gimp_channel >= 0) {
				if (old_width > 0 && old_height > 0) {
					bitmap->width = old_width;
					bitmap->height = old_height;
					free_layer_data(bitmap, layer);
				}
				bitmap->width = new_width;
				bitmap->height = new_height;
				allocate_layer_data(bitmap, layer);
			}
		}
	}
	bitmap->width = new_width;
	bitmap->height = new_height;
}

void remove_bitmap(i_bitmap_t *bitmap)
{
	GList *i;
	i_layer_t *layer;

	if (bitmap->layers != NULL) {
		for (i = bitmap->layers; i != NULL; i = g_list_next(i)) {
			layer = (i_layer_t*)i->data;
			g_free(layer->name);
			free_layer_data(bitmap, layer);
		}
		g_list_free(bitmap->layers);
		if (bitmap->redraw_queue != NULL) {
			g_array_free(bitmap->redraw_queue, TRUE);
			bitmap->redraw_queue = NULL;
		}
	}
	g_free(bitmap->name);
	vg->bitmaps = g_list_remove(vg->bitmaps, bitmap);
}

void unlink_bitmap(i_bitmap_t *bitmap)
{
	GList *i;
	i_layer_t *layer;

	bitmap->download = FALSE;
	bitmap->upload = FALSE;
	bitmap->gimp_image_id = 0;
	bitmap->gimp_layer_id = 0;

	if (bitmap->layers != NULL) {
		for (i = bitmap->layers; i != NULL; i = g_list_next(i)) {
			layer = (i_layer_t*)i->data;
			if (layer->gimp_channel >= 0) {
				verse_send_b_layer_unsubscribe(bitmap->verse_node_id, layer->verse_layer_id);
			}
		}
	}
	gui_refresh_verse_list();
}

void remove_layer(i_layer_t *layer)
{
	g_free(layer->name);
	free_layer_data(layer->parent, layer);
	layer->parent->layers = g_list_remove(layer->parent->layers, layer);
}

/* ### GTK Callbacks ### */

static void cb_gui_connect(GtkWidget *widget, gpointer data)
{
	vrs_connect((char*)gtk_entry_get_text(GTK_ENTRY(vg->address)), (char*)gtk_entry_get_text(GTK_ENTRY(vg->uname)), (char*)gtk_entry_get_text(GTK_ENTRY(vg->pass)));
	/*vg->verse_timeout = gtk_idle_add(vrs_callback_update, NULL);*/
	vg->verse_timeout = gtk_timeout_add(100, vrs_callback_update, NULL);
}

static void cb_gui_disconnect(GtkWidget *widget, gpointer data)
{
	vrs_disconnect(vg->session, (char*)gtk_entry_get_text(GTK_ENTRY(vg->address)), "Gimp Verse Interaction Ends.");
}

static void cb_gui_upload(GtkWidget *widget, gpointer data)
{
	GtkTreeSelection *select;
	GtkTreeIter iter;
	GtkTreeModel *model;
	gint32 id;
	uint i, j;

	select = gtk_tree_view_get_selection (GTK_TREE_VIEW(vg->gmptree));
	if (gtk_tree_selection_get_selected (select, &model, &iter))
		gtk_tree_model_get (model, &iter, 0, &id, -1);
	else 
		return;
	if (get_bitmap_verse_id(id) != NULL) {
		gimp_message("Bitmap already uploaded.\n");
		return;
	}
	for (i = 0; (j = g_array_index(vg->upload_queue, guint, i)) != 0; i++) {
		if (j == id)
			return;
	}
	g_array_append_val(vg->upload_queue, id);
	vrs_create_bitmap();
}

static i_bitmap_t *gui_get_selected_bitmap()
{
	GtkTreeSelection *select;
	GtkTreeIter m_iter;
	GtkTreeModel *model;
	gint32 id;

	select = gtk_tree_view_get_selection (GTK_TREE_VIEW(vg->vrstree));
	if (gtk_tree_selection_get_selected (select, &model, &m_iter))
		gtk_tree_model_get (model, &m_iter, 0, &id, -1);
	else
		return NULL;

	return get_bitmap(id);
}

static void cb_gui_destroy(GtkWidget *widget, gpointer data)
{
	i_bitmap_t *bitmap;
	
	if ((bitmap = gui_get_selected_bitmap()) != NULL)
		verse_send_node_destroy(bitmap->verse_node_id);
}

static void cb_gui_download(GtkWidget *widget, gpointer data)
{
	i_bitmap_t *bitmap;
	GList *i;
	i_layer_t *layer;
	
	if ((bitmap = gui_get_selected_bitmap()) == NULL)
		return;
	if (!create_gimp_image_from_verse(bitmap)) {
		gimp_message("Already subscribed to bitmap.\n");
		return;
	}
	if (bitmap->width == 0 || bitmap->height == 0) {
		gimp_message("Can't subscribe, invalid bitmap dimensions.\n");
		return;
	}
	bitmap->download = TRUE;
	bitmap->upload = FALSE;

	if (bitmap->layers != NULL)
		for (i = bitmap->layers; i != NULL; i = g_list_next(i)) {
			layer = (i_layer_t*)i->data;
			if (layer->gimp_channel>=0) {
				free_layer_data(bitmap, layer);
				allocate_layer_data(bitmap, layer);
				verse_send_b_layer_subscribe(bitmap->verse_node_id, layer->verse_layer_id, 0);
			}
		}

	gui_refresh_verse_list();
	gui_refresh_gimp_list();
}

static void cb_gui_unlink(GtkWidget *widget, gpointer data)
{
	i_bitmap_t *bitmap;
	
	if ((bitmap = gui_get_selected_bitmap()) == NULL)
		return;
	unlink_bitmap(bitmap);
}

static void cb_gui_mode(GtkWidget *widget, gpointer data)
{
	i_bitmap_t *bitmap;

	if ((bitmap = gui_get_selected_bitmap()) == NULL)
		return;

	bitmap->download = !(bitmap->download);
	bitmap->upload = !(bitmap->upload);

	gui_refresh_verse_list();
}

static void cb_gui_speed_changed(GtkAdjustment *adj)
{
	if (vg->gimp_timeout) {
		gtk_timeout_remove(vg->gimp_timeout);
		vg->gimp_timeout = gtk_timeout_add(gtk_adjustment_get_value(GTK_ADJUSTMENT(vg->speedadj)), gimp_images_update, NULL);
	}
}

static void gui_refresh_gimp_list()
{
	gchar *name;
	gint32 *images;
	gint num_images;
	gint i;
	GtkTreeIter iter;

	images = gimp_image_list(&num_images);
	gtk_list_store_clear(vg->gmplist);
	for (i = 0; i < num_images; i++)
	{
		name = gimp_image_get_name(images[i]);
		gtk_list_store_append (vg->gmplist, &iter);
		gtk_list_store_set (vg->gmplist, &iter, 0, images[i], 1, name, -1);
		g_free(name);
	}
	g_free(images);
}

void gui_refresh_verse_list()
{
	GList *i;
	GtkTreeIter tree_iter;
	i_bitmap_t *bitmap;
	gchar *name, *mode;

	gtk_list_store_clear(vg->vrslist);
	for (i = vg->bitmaps; i != NULL; i = g_list_next(i)) {
		bitmap = (i_bitmap_t*)i->data;
		gtk_list_store_append (vg->vrslist, &tree_iter);

		if (bitmap->download==TRUE)
			mode = g_strdup("download");
		else if (bitmap->upload==TRUE)
			mode = g_strdup("upload");
		else
			mode = g_strdup("n/a");

		if (bitmap->name != NULL)
			name = g_strdup_printf("%s (%ix%i)", bitmap->name, bitmap->width, bitmap->height);
		else
			name = g_strdup_printf("*** (%ix%i)", bitmap->width, bitmap->height);
		gtk_list_store_set (vg->vrslist, &tree_iter, 0, bitmap->verse_node_id, 1, name, 2, mode, 3, bitmap->gimp_image_id, -1);
		g_free(name);
		g_free(mode);
	}
}

/* ### Program Init, Quit, and GUI Building ### */

static void nice_quit()
{
	if (vg->connected||vg->connecting) {
		vrs_disconnect(vg->session, (char*)gtk_entry_get_text(GTK_ENTRY(vg->address)), "Gimp Verse Interaction Ends.");
		gtk_timeout_add(500, gui_quit, NULL);
	} else if (vg->disconnecting)
		gtk_timeout_add(500, gui_quit, NULL);
	else
		gui_quit(NULL);
}

static gboolean gui_quit(gpointer user)
{
	gtk_widget_destroy(vg->dialog);
	return FALSE;
}

static void gui_create_server_data()
{
	vg->upload_queue = g_array_new(TRUE, FALSE, sizeof(gint32));
	gui_refresh_gimp_list();
	vg->gimp_timeout = gtk_timeout_add(gtk_adjustment_get_value(GTK_ADJUSTMENT(vg->speedadj)), gimp_images_update, NULL);
}

void gui_destroy_server_data()
{
	GList *i, *j;
	i_bitmap_t *bitmap;
	i_layer_t *layer;

	if (vg->gimp_timeout) {
		gtk_timeout_remove(vg->gimp_timeout);
		vg->gimp_timeout = 0;
	}
	if (vg->bitmaps != NULL) {
		for (i = vg->bitmaps; i != NULL; i = g_list_next(i)) {
			bitmap = (i_bitmap_t*)i->data;
			if (bitmap->layers != NULL) {
				for (j = bitmap->layers; j != NULL; j = g_list_next(j)) {
					layer = (i_layer_t*)j->data;
					g_free(layer->name);
					layer->name = NULL;
					free_layer_data(bitmap, layer);
					layer->data = NULL;
				}
				g_list_free(bitmap->layers);
				if (bitmap->redraw_queue != NULL) {
					g_array_free(bitmap->redraw_queue, TRUE);
					bitmap->redraw_queue = NULL;
				}
			}
		}
		g_free(bitmap->name);
		bitmap->name = NULL;
		g_list_free(vg->bitmaps);
		vg->bitmaps = NULL;
	}

	if (vg->upload_queue != NULL) {
		g_array_free(vg->upload_queue, TRUE);
		vg->upload_queue = NULL;
	}
	gui_refresh_verse_list();
}

static void gui_destroy()
{	
	gui_destroy_server_data();
	if (vg->connected||vg->connecting)
		vrs_disconnect(vg->session, (char*)gtk_entry_get_text(GTK_ENTRY(vg->address)), "Gimp Verse Interaction Ends.");
	gtk_main_quit();
}

void gui_init()
{
	GtkWidget *vbox, *vbox2, *hbox, *label, *mode, *hscale;
	GtkWidget *unlink, *upload, *refresh_gimp, *destroy, *topvbox, *scwin;

	/* The Dialog and notebook. */
	vg = g_new0(verse_gimp_t, 1);
	gimp_ui_init("verse_gimp", FALSE);
	vg->dialog = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(vg->dialog), "Verse Gimp Interaction");
	gtk_window_set_role(GTK_WINDOW(vg->dialog), "gimp-1.3");
	topvbox = gtk_vbox_new(FALSE, 6);
	gtk_container_add(GTK_CONTAINER(vg->dialog), topvbox);
	
	g_signal_connect(vg->dialog, "destroy", G_CALLBACK (gui_destroy), vg);
	vg->notebook = gtk_notebook_new();
	gtk_box_pack_start(GTK_BOX(topvbox), vg->notebook, TRUE, TRUE, 0);

	/* Connection Settings */
	vbox2 = gtk_vbox_new(FALSE, 3);

	vg->table_server = gtk_table_new(4, 2, FALSE);
	gtk_table_set_col_spacings (GTK_TABLE (vg->table_server), 4);
	gtk_table_set_row_spacings (GTK_TABLE (vg->table_server), 2);
	gtk_container_set_border_width (GTK_CONTAINER (vg->table_server), 3);

	label = gtk_label_new ("Server:"); 
	gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
	gtk_table_attach_defaults(GTK_TABLE(vg->table_server), label,
	                         0, 1, 0, 1);
	vg->address = gtk_entry_new();
	gtk_entry_set_text(GTK_ENTRY(vg->address), "localhost");
	gtk_table_attach_defaults(GTK_TABLE(vg->table_server), vg->address,
	                         1, 2, 0, 1);

	label = gtk_label_new ("Username:"); 
	gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
	gtk_table_attach_defaults(GTK_TABLE(vg->table_server), label,
	                         0, 1, 2, 3);
	vg->uname = gtk_entry_new();
	gtk_entry_set_text(GTK_ENTRY(vg->uname), "anonymous");
	gtk_table_attach_defaults(GTK_TABLE(vg->table_server), vg->uname,
	                         1, 2, 2, 3);

	label = gtk_label_new ("Password:"); 
	gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
	gtk_table_attach_defaults(GTK_TABLE(vg->table_server), label,
	                         0, 1, 3, 4);
	vg->pass = gtk_entry_new();
	gtk_entry_set_text(GTK_ENTRY(vg->pass), "anonymous");
	gtk_table_attach_defaults(GTK_TABLE(vg->table_server), vg->pass,
	                         1, 2, 3, 4);

	gtk_box_pack_start(GTK_BOX(vbox2), vg->table_server, FALSE, FALSE, 3);
	
	vg->state_label = gtk_label_new("");
	gtk_box_pack_start(GTK_BOX (vbox2), vg->state_label, FALSE, FALSE, 0);

	hbox = gtk_hbox_new(FALSE, 3);
	label = gtk_label_new ("Update interval:"); 
	gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 3);
	vg->speedadj = gtk_adjustment_new(100, 20, 1000, 10, 0, 0);
	hscale = gtk_hscale_new(GTK_ADJUSTMENT (vg->speedadj));
	gtk_box_pack_start(GTK_BOX(hbox), hscale, TRUE, TRUE, 3);
	g_signal_connect(vg->speedadj, "value_changed", G_CALLBACK(cb_gui_speed_changed), NULL);

	gtk_box_pack_start(GTK_BOX(vbox2), hbox, FALSE, FALSE, 3);


	label = gtk_label_new("Connection");
	gtk_notebook_append_page (GTK_NOTEBOOK(vg->notebook), vbox2, label);

	/* The Gimp Images List tab */
	vbox = gtk_vbox_new(FALSE, 3);
	vg->gmplist = gtk_list_store_new (2, G_TYPE_INT, G_TYPE_STRING);
	vg->gmptree = gtk_tree_view_new_with_model(GTK_TREE_MODEL(vg->gmplist));
	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(vg->gmptree),
	                                           -1, "Gimp ID",
						   gtk_cell_renderer_text_new(),
						   "text", 0, NULL);
	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(vg->gmptree),
	                                           -1, "Name",
						   gtk_cell_renderer_text_new(),
						   "text", 1, NULL);
	scwin = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW(scwin),
					GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_container_add (GTK_CONTAINER( scwin), vg->gmptree);
	gtk_box_pack_start (GTK_BOX (vbox), scwin, TRUE, TRUE, 0);

	hbox = gtk_hbox_new(FALSE, 3);
	upload = gtk_button_new_with_mnemonic("Upload to Verse Server");
	g_signal_connect(upload, "clicked", G_CALLBACK(cb_gui_upload), NULL);
	gtk_box_pack_start (GTK_BOX (hbox), upload, TRUE, TRUE, 0);

	refresh_gimp = gtk_button_new_with_mnemonic("Refresh List");
	g_signal_connect(refresh_gimp, "clicked", G_CALLBACK(gui_refresh_gimp_list), NULL);
	gtk_box_pack_start (GTK_BOX (hbox), refresh_gimp, TRUE, TRUE, 0);
	gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);

	/* The Verse Bitmaps List tab */
	vg->vrslist = gtk_list_store_new (4, G_TYPE_INT, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT);
	vg->vrstree = gtk_tree_view_new_with_model(GTK_TREE_MODEL(vg->vrslist));
	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(vg->vrstree),
	                                           -1, "Verse ID",
						   gtk_cell_renderer_text_new(),
						   "text", 0, NULL);
	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(vg->vrstree),
	                                           -1, "Name",
						   gtk_cell_renderer_text_new(),
						   "text", 1, NULL);
	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(vg->vrstree),
	                                         -1, "Mode",
						 gtk_cell_renderer_text_new(),
				       "text", 2, NULL);
	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(vg->vrstree),
	                                           -1, "Gimp ID",
						   gtk_cell_renderer_text_new(),
						   "text", 3, NULL);
	scwin = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW(scwin),
					GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_container_add (GTK_CONTAINER (scwin), vg->vrstree);
	gtk_box_pack_start (GTK_BOX (vbox), scwin, TRUE, TRUE, 0);

	hbox = gtk_hbox_new(FALSE, 3);
	upload = gtk_button_new_with_mnemonic("Subscribe");
	g_signal_connect(upload, "clicked", G_CALLBACK(cb_gui_download), NULL);
	gtk_box_pack_start (GTK_BOX (hbox), upload, TRUE, TRUE, 0);

	unlink = gtk_button_new_with_mnemonic("Unsubscribe");
	g_signal_connect(unlink, "clicked", G_CALLBACK(cb_gui_unlink), NULL);
	gtk_box_pack_start (GTK_BOX (hbox), unlink, TRUE, TRUE, 0);

	mode = gtk_button_new_with_mnemonic("Switch Mode");
	g_signal_connect(mode, "clicked", G_CALLBACK(cb_gui_mode), NULL);
	gtk_box_pack_start (GTK_BOX (hbox), mode, TRUE, TRUE, 0);

	destroy = gtk_button_new_with_mnemonic("Destroy");
	g_signal_connect(destroy, "clicked", G_CALLBACK(cb_gui_destroy), NULL);
	gtk_box_pack_start (GTK_BOX (hbox), destroy, TRUE, TRUE, 0);
	gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);

	label = gtk_label_new ("Images");
	gtk_notebook_append_page (GTK_NOTEBOOK(vg->notebook), vbox, label);

	hbox = gtk_hbox_new(FALSE, 3);
	vg->conn = gtk_button_new_with_label("Connect");
	g_signal_connect(vg->conn, "clicked", G_CALLBACK(cb_gui_connect), NULL);
	gtk_box_pack_start(GTK_BOX(hbox), vg->conn, TRUE, TRUE, 0);
	vg->disconn = gtk_button_new_with_label("Disconnect");
	g_signal_connect(vg->disconn, "clicked", G_CALLBACK(cb_gui_disconnect), NULL);
	gtk_box_pack_start(GTK_BOX(hbox), vg->disconn, TRUE, TRUE, 0);
	destroy = gtk_button_new_from_stock(GTK_STOCK_CLOSE);
	g_signal_connect(destroy, "clicked", G_CALLBACK(nice_quit), NULL);
	gtk_box_pack_start(GTK_BOX(hbox), destroy, TRUE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(topvbox), hbox, FALSE, FALSE, 0);

	gtk_widget_set_size_request(vg->dialog, 300, 300);
	gtk_widget_show_all(vg->dialog);
	gui_set_state_disconnected();

	gtk_main();
	gdk_flush();
}

void gui_set_state_connecting()
{
	vg->connected = FALSE;
	vg->connecting = TRUE;
	vg->disconnecting = FALSE;
	gtk_widget_set_sensitive(vg->table_server, FALSE);
	gtk_widget_set_sensitive(vg->conn, FALSE);
	gtk_widget_set_sensitive(vg->disconn, TRUE);
	gtk_widget_set_sensitive(gtk_notebook_get_nth_page(GTK_NOTEBOOK(vg->notebook), 1), FALSE);
	gtk_label_set_markup(GTK_LABEL(vg->state_label), "State: <b>connecting</b>");
	
}

void gui_set_state_connected()
{
	vg->connected = TRUE;
	vg->connecting = FALSE;
	vg->disconnecting = FALSE;
	gtk_widget_set_sensitive(vg->table_server, FALSE);
	gtk_widget_set_sensitive(vg->conn, FALSE);
	gtk_widget_set_sensitive(vg->disconn, TRUE);
	gtk_widget_set_sensitive(gtk_notebook_get_nth_page(GTK_NOTEBOOK(vg->notebook), 1), TRUE);
	gtk_label_set_markup(GTK_LABEL(vg->state_label), "State: <b>connected</b>");

	gui_create_server_data();
}

void gui_set_state_disconnected()
{
	vg->connected = FALSE;
	vg->connecting = FALSE;
	vg->disconnecting = FALSE;
	gtk_widget_set_sensitive(vg->table_server, TRUE);
	gtk_widget_set_sensitive(vg->conn, TRUE);
	gtk_widget_set_sensitive(vg->disconn, FALSE);
	gtk_notebook_set_current_page(GTK_NOTEBOOK(vg->notebook), 0);
	gtk_widget_set_sensitive(gtk_notebook_get_nth_page(GTK_NOTEBOOK(vg->notebook), 1), FALSE);
	gtk_label_set_markup(GTK_LABEL(vg->state_label), "State: <b>disconnected</b>");
}

void gui_set_state_disconnecting()
{
	/* FIXME */
	gui_set_state_disconnected();
}

