
#include <string.h>

#include "verse_gimp.h"

/* Callbacks */

static void cb_connect_accept(void *user, VNodeID avatar, const char *addr, uint8 *hid)
{
	gui_set_conn_state(CONNECTED);

	/* verse gimp will receive changes of bitmap and object nodes, because
	 * one object node represents gimp as verse avatar */
	verse_send_node_index_subscribe(1 << V_NT_BITMAP | 1 << V_NT_OBJECT);
}

static void cb_node_create(void *user, VNodeID node_id, uint8 type, VNodeOwner owner)
{
	if (type == V_NT_BITMAP) {
		sync_verse_bitmap_create(node_id, owner);
	}
	else if ((type == V_NT_OBJECT) && (owner == VN_OWNER_MINE)) {
		gchar *base_name = "gimp_";
		gchar *name = NULL;
		gchar *version = gimp_version();

		if(version) {
			printf("base_name: %s\n", base_name);
			printf("version: %s\n", version);
			name = malloc(sizeof(gchar)*(strlen(base_name) + strlen(version)));
			name[0] = '\0';
			strcat(name, base_name);
			strcat(name, version);
		}

		verse_send_node_name_set(node_id, name);
		if(name) free(name);
	}
}

static void cb_node_destroy(void *user, VNodeID node_id)
{
	VGBitmap *bmp = sync_bitmap_lookup_verse(node_id);

	if (bmp)
		sync_verse_bitmap_destroy(bmp);
}

static void cb_node_name_set(void *user, VNodeID node_id, const char *name)
{
	VGBitmap *bmp = sync_bitmap_lookup_verse(node_id);

	if (bmp)
		sync_verse_bitmap_name_set(bmp, name);
}

static void cb_dimensions(void *user_data, VNodeID node_id, uint16 width, uint16 height, uint16 depth)
{
	VGBitmap *bmp = sync_bitmap_lookup_verse(node_id);

	if (bmp)
		sync_verse_bitmap_dimensions(bmp, width, height);
}

static void cb_layer_create(void *user_data, VNodeID node_id, VLayerID layer_id, const char *name, VNBLayerType type)
{
	VGBitmap *bmp = sync_bitmap_lookup_verse(node_id);

	if (bmp)
		sync_verse_layer_create(bmp, layer_id, name);
}

static void cb_layer_destroy(void *user_data, VNodeID node_id, VLayerID layer_id)
{
	VGBitmap *bmp = sync_bitmap_lookup_verse(node_id);

	if (bmp) {
		int channel;
		VGLayer *layer = sync_layer_lookup_verse(bmp, layer_id, &channel);

		if (layer)
			sync_verse_layer_destroy(bmp, layer, channel);
	}
}

static void cb_tile(void *user_data, VNodeID node_id, VLayerID layer_id, uint16 tile_x, uint16 tile_y, uint16 z, VNBLayerType type, VNBTile *data)
{
	VGBitmap *bmp = sync_bitmap_lookup_verse(node_id);

	if (bmp) {
		int channel;
		VGLayer *layer = sync_layer_lookup_verse(bmp, layer_id, &channel);

		if (layer) {
			int i, j;
			uint8 tile[VN_B_TILE_SIZE*VN_B_TILE_SIZE];

			if (type == VN_B_LAYER_UINT1) {
				for (i = 0; i < VN_B_TILE_SIZE; i++)
					for (j = 0; j < VN_B_TILE_SIZE; j++)
						tile[i + j*VN_B_TILE_SIZE] = (data->vuint1[j] & (1 << i))*255;
			} else if (type == VN_B_LAYER_UINT8) {
				for (i = 0; i < VN_B_TILE_SIZE*VN_B_TILE_SIZE; i++)
					tile[i] = data->vuint8[i]; 
			} else if (type == VN_B_LAYER_UINT16) {
				for (i = 0; i < VN_B_TILE_SIZE*VN_B_TILE_SIZE; i++)
					tile[i] = data->vuint16[i] / 255;
			} else if (type == VN_B_LAYER_REAL32) {
				for (i = 0; i < VN_B_TILE_SIZE*VN_B_TILE_SIZE; i++)
					tile[i] = data->vreal32[i] * 255;
			} else if (type == VN_B_LAYER_REAL64) {
				for (i = 0; i < VN_B_TILE_SIZE*VN_B_TILE_SIZE; i++)
					tile[i] = data->vreal64[i] * 255;
			}

			sync_verse_layer_tile_set(layer, channel, tile_x, tile_y, tile);
		}
	}
}

/* Connection */

void conn_start(const char *addr, const char *uname, const char *pass)
{
	verse_callback_set(verse_send_connect_accept, cb_connect_accept, NULL);
	verse_callback_set(verse_send_node_create, cb_node_create, NULL);
	verse_callback_set(verse_send_node_destroy, cb_node_destroy, NULL);
	verse_callback_set(verse_send_node_name_set, cb_node_name_set, NULL);
	verse_callback_set(verse_send_b_dimensions_set, cb_dimensions, NULL);
	verse_callback_set(verse_send_b_layer_create, cb_layer_create, NULL);
	verse_callback_set(verse_send_b_layer_destroy, cb_layer_destroy, NULL);
	verse_callback_set(verse_send_b_tile_set, cb_tile, NULL);

	G.conn.session = verse_send_connect(uname, pass, addr, NULL);
	verse_session_set(G.conn.session);
}

void conn_stop()
{
	verse_send_connect_terminate(NULL, "Gimp plug-in quit.");
	verse_callback_update(10000);
	verse_session_destroy(G.conn.session);
	G.conn.session = NULL;
}

void conn_callback_update()
{
	verse_callback_update(10000);
}

